from enum import Enum
from typing import List, Set


class ShellType(str, Enum):
    """An enum with the supported shell types for tab autocompletion."""

    BASH = "bash"
    ZSH = "zsh"
    FISH = "fish"

    @classmethod
    def get_all(cls) -> Set[str]:
        """Return a set of all the supported shell types."""

        return {e for e in cls}

    @classmethod
    def get_all_values(cls) -> List[str]:
        """Return a list with all the enum members' values, the shell names."""

        return [e.value for e in cls]



class AutoClickAutoError(Exception):
    """Custom exception with an error message for `auto_click_auto`."""

    def __init__(self, message):
        super().__init__(message)


class ShellConfigurationFileNotFoundError(AutoClickAutoError):
    """Exception raised when the shell configuration file does not exist."""

    pass


class ShellTypeNotSupportedError(AutoClickAutoError):
    """Exception raised when the shell type is not supported."""

    pass


class ShellEnvVarNotFoundError(AutoClickAutoError):
    """
    Exception raised when the shell type cannot be inferred from the `SHELL`
    environment variable.
    """

    pass



import os
import platform
from typing import TYPE_CHECKING, Any, Callable, Optional, Set, TypeVar, Union

from click import Command, Context, Parameter, option

from .constants import ShellType
from .exceptions import ShellEnvVarNotFoundError, ShellTypeNotSupportedError
from .utils import add_shell_configuration, create_file, detect_shell

if TYPE_CHECKING:
    import typing_extensions as te

R = TypeVar("R")
T = TypeVar("T")
_AnyCallable = Callable[..., Any]
_Decorator: "te.TypeAlias" = Callable[[T], T]
FC = TypeVar("FC", bound=Union[_AnyCallable, Command])


def enable_click_shell_completion(
    program_name: str,
    shells: Optional[Set[ShellType]] = None,
    verbose: Optional[bool] = False,
) -> None:
    """
    Enable tab completion for Click's supported shell types.

    If ``shells`` is not provided, `auto-click-auto` will attempt to detect the
    shell type the user is currently running the program on.

    See https://click.palletsprojects.com/en/latest/shell-completion.
    `auto-click-auto` is using the `eval` command implementation suggested from
    Click.

    :param program_name: The program name for which we enable shell completion.
    :param shells: The shell types for which we want to add tab completion
    support.
    :param verbose: `True` to print more details regarding the enabling,
    `False` otherwise. If this function is called on every run of the CLI
    program it might be better to set to `False`.
    :raise NotImplementedError: When ``shells`` option is not supported.
    """

    # Check that the program is run on one of the supported Operating Systems
    supported_os = ("Linux", "MacOS")
    os_name = platform.system()
    if os_name not in supported_os:
        if verbose is True:
            print(
                f"{os_name} is not one of the supported Operating Systems "
                f"({supported_os}) of `auto-click-auto`."
            )

        return None

    click_env_var = f"_{program_name.upper()}_COMPLETE"

    if shells is None:
        try:
            shells = {detect_shell()}

        except (ShellTypeNotSupportedError, ShellEnvVarNotFoundError) as err:
            if verbose is True:
                print(err)

            return None

    for shell in shells:

        if shell in (ShellType.BASH, ShellType.ZSH):
            shell_config_file = os.path.expanduser(f"~/.{shell}rc")

            # Completion implementation: `eval` command in shell configuration
            eval_command = (
                f'eval \"$({click_env_var}={shell.value}_source '
                f'{program_name})\"'
            )
            add_shell_configuration(
                shell_config_file=shell_config_file,
                config_string=eval_command,
                verbose=verbose,
            )

        elif shell == ShellType.FISH:
            completer_script_path = os.path.expanduser(
                f"~/.config/fish/completions/{program_name}.{shell}"
            )

            # bash and zsh config files are generic, so we can assume the user
            # already has created them. fish's shell configuration file for
            # custom completions is specific to the program name, so we will
            # create it if it doesn't already exist.
            create_file(file_path=completer_script_path)

            # Completion implementation: `eval` command in shell configuration
            eval_command = (
                f"eval (env {click_env_var}={shell.value}_source "
                f"{program_name})"
            )
            add_shell_configuration(
                shell_config_file=completer_script_path,
                config_string=eval_command,
                verbose=verbose,
            )

        else:
            raise NotImplementedError


def enable_click_shell_completion_option(
    *param_decls: str,
    program_name: Optional[str] = None,
    shells: Optional[Set[ShellType]] = None,
    **kwargs: Any,
) -> _Decorator[FC]:
    """
    Add a ``--autocomplete`` option which enables tab completion and exits the
    program. This function can be used as a decorator in a Click command.

    If ``program_name`` is not provided, it will be detected from the command.

    If ``shells`` is not provided, `auto-click-auto` will attempt to detect the
    shell type the user is currently running the program on.

    Uses Click's `option` function. It is possible to pass the relevant
    function arguments to override the defaults ones, e.g., ``expose_value``,
    ``help``, etc.

    :param param_decls: One or more option names. Defaults to the single value
    ``"--autocomplete"``.
    :param program_name: The program name for which we enable shell completion.
    :param shells: The shell types for which we want to add tab completion
    support.
    :param kwargs: Extra arguments are passed to :func:`option`.
    """

    def callback(ctx: Context, param: Parameter, value: bool) -> None:
        if not value or ctx.resilient_parsing:
            return

        nonlocal program_name
        nonlocal shells

        if program_name is None:
            program_name = ctx.find_root().info_name

        assert program_name is not None

        enable_click_shell_completion(
            program_name=program_name, shells=shells, verbose=True
        )

        ctx.exit()
        return None

    if not param_decls:
        param_decls = ("--autocomplete",)

    kwargs.setdefault("is_flag", True)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("help", "Enable tab autocompletion and exit.")
    kwargs["callback"] = callback

    return option(*param_decls, **kwargs)



import os
from typing import List, Optional

from auto_click_auto.constants import ShellType
from auto_click_auto.exceptions import (
    ShellConfigurationFileNotFoundError,
    ShellEnvVarNotFoundError,
    ShellTypeNotSupportedError,
)


def create_file(file_path: str) -> None:
    """
    Check if the file and the directories of the given file path exist and if
    not create them.

    :param file_path: The path of the file the check and creation is done for.
    :return: `None`
    """

    if not os.path.exists(file_path):
        directory = os.path.dirname(file_path)

        if not os.path.exists(directory):
            # Recursive directory creation
            os.makedirs(directory)

        # Open for exclusive creation, failing if the file already exists
        with open(file_path, 'x'):
            pass


def check_strings_in_file(file_path: str, search_strings: List[str]) -> bool:
    """
    Check if the given search strings are in the specified file.

    :param file_path: The path of the file the search is done for.
    :param search_strings: The strings that are searched for in the file.
    :return: `True`, if any of the strings are found in the file, and `False`
    otherwise.
    """

    try:
        with open(file_path) as file:
            return any(
                search_string in line for line in file
                for search_string in search_strings
            )

    except FileNotFoundError:
        raise ShellConfigurationFileNotFoundError(
            f"The {file_path} configuration file does not exist."
        )


def add_shell_configuration(
        shell_config_file: str,
        config_string: str,
        verbose: Optional[bool] = False
) -> None:
    """
    Add the given configuration in the specified shell configuration file, if
    the string is not already present.

    :param shell_config_file: The shell configuration file to add the string
    to.
    :param config_string: The desired configuration string.
    :param verbose: `True` to print whether the configuration is already in the
    file, `False` otherwise.
    """

    try:
        string_in_file = check_strings_in_file(
                file_path=shell_config_file, search_strings=[config_string]
        )
    except ShellConfigurationFileNotFoundError as err:
        if verbose is True:
            print(err)

        return None

    if not string_in_file:
        print(
            f"Adding tab autocomplete configuration in {shell_config_file} ..."
        )

        os.system(
            "echo '# Shell completion configuration for the Click Python "
            f"package' >> {shell_config_file}"
        )
        # Source the file in the shell config file
        os.system(f"echo '{config_string}' >> {shell_config_file}")

        print(
            "Restart or create a new shell session for the changes to take "
            "effect."
        )

    elif string_in_file and verbose is True:
        print(
            "Tab autocomplete configuration already setup in "
            f"{shell_config_file}."
        )


def detect_shell() -> ShellType:
    """
    Attempt to detect the shell type using the `SHELL` environment variable.

    :raise ShellEnvVarNotFoundError: When the `SHELL` environment variable is
    not set in the system.
    :raise ShellTypeNotSupportedError: When the shell value returned from the
    `SHELL` environment variable does not belong to one of the supported
    shells.
    """

    try:
        shell_env_var = os.environ["SHELL"]

    except KeyError:
        raise ShellEnvVarNotFoundError(
            "Could not infer the shell type from the 'SHELL' environment "
            "variable."
        )

    try:
        shell_value = shell_env_var.split("/")[-1]
        return ShellType(shell_value)

    except ValueError:
        raise ShellTypeNotSupportedError(
            f"{shell_value} is not one of the supported shell types "
            f"({ShellType.get_all_values()})."
        )
